import time
from typing import Optional
from asyncworker.rabbitmq.message import RabbitMQMessage
from asyncworker.metrics.registry import REGISTRY
from barterdude.hooks import BaseHook
from barterdude.hooks.metrics.prometheus.definitions import Definitions
from barterdude.hooks.metrics.prometheus.metrics import Metrics
from prometheus_client import CollectorRegistry


class Prometheus(BaseHook):

    def __init__(
        self,
        labels: dict = {},
        registry: CollectorRegistry = None
    ):
        self.__registry = registry or REGISTRY
        self.__labels = labels
        self.__metrics = Metrics(self.__registry)
        self.__definitions = Definitions(
            self.__registry, self.__metrics, list(self.__labels.keys())
        )
        self._msg_start = {}
        self.__definitions.save_metrics()

    @property
    def metrics(self):
        return self.__metrics

    async def before_consume(self, message: RabbitMQMessage):
        hash_message = id(message)
        self._msg_start[hash_message] = time.time()
        metric = self.metrics[self.__definitions.BEFORE_CONSUME]
        if self.__labels:
            metric = metric.labels(**self.__labels)
        metric.inc()

    async def _on_complete(self,
                           message: RabbitMQMessage,
                           state: str,
                           error: Optional[Exception] = None):

        final_time = time.time()
        hash_message = id(message)
        labels = self.__labels.copy()
        labels["state"] = state
        labels["error"] = str(type(error)) if error else ""
        self.metrics[self.__definitions.HISTOGRAM_MEASURE].labels(
            **labels).observe(
                final_time - self._msg_start.pop(hash_message)
        )

    async def on_success(self, message: RabbitMQMessage):
        await self._on_complete(message, self.__definitions.SUCCESS)

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        await self._on_complete(message, self.__definitions.FAIL, error)

    async def on_connection_fail(self, error: Exception, retries: int):
        metric = self.metrics[self.__definitions.CONNECTION_FAIL]
        if self.__labels:
            metric = metric.labels(**self.__labels)
        metric.inc()



import logging
import os
from pythonjsonlogger import jsonlogger

BARTERDUDE_DEFAULT_APP_NAME = os.environ.get(
    "BARTERDUDE_DEFAULT_APP_NAME", "barterdude"
)
BARTERDUDE_DEFAULT_LOG_LEVEL = int(
    os.environ.get(
        "BARTERDUDE_DEFAULT_LOG_LEVEL", logging.INFO
    )
)
BARTERDUDE_LOG_REDACTED = bool(int(os.environ.get(
    "BARTERDUDE_LOG_REDACTED", "1"
)))
handler = logging.StreamHandler()
handler.setFormatter(jsonlogger.JsonFormatter(
    '%(levelname)s %(name)s %(pathname)s %(lineno)s',
    timestamp=True
))
default_logger = logging.getLogger(BARTERDUDE_DEFAULT_APP_NAME)


def getLogger(name, level=BARTERDUDE_DEFAULT_LOG_LEVEL):
    logger = default_logger.getChild(name)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger



from aiohttp import web
from abc import ABCMeta, abstractmethod
from barterdude import BarterDude
from asyncworker.rabbitmq.message import RabbitMQMessage


class BaseHook(metaclass=ABCMeta):
    @abstractmethod
    async def on_success(self, message: RabbitMQMessage):
        '''Called after successfuly consumed the message'''

    @abstractmethod
    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        '''Called when fails to consume the message'''

    @abstractmethod
    async def before_consume(self, message: RabbitMQMessage):
        '''Called before consuming the message'''

    @abstractmethod
    async def on_connection_fail(self, error: Exception, retries: int):
        '''Called when the consumer fails to connect to the broker'''


class HttpHook(BaseHook):
    def __init__(
            self,
            barterdude: BarterDude,
            path: str,
            method: str = "GET"):

        barterdude.add_endpoint(
            routes=[path],
            methods=[method],
            hook=self.__call__
        )

    async def __call__(self, req: web.Request):
        raise NotImplementedError

    async def on_success(self, message: RabbitMQMessage):
        raise NotImplementedError

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        raise NotImplementedError

    async def before_consume(self, message: RabbitMQMessage):
        raise NotImplementedError

    async def on_connection_fail(self, error: Exception, retries: int):
        raise NotImplementedError



from abc import ABC, abstractmethod
import json

from barterdude import BarterDude
from barterdude.hooks import HttpHook
from asyncworker.rabbitmq.message import RabbitMQMessage
from aiohttp import web
from time import time
from collections import deque
from bisect import bisect_left


def _remove_old(instants: deque, old_timestamp: float):
    pos = bisect_left(instants, old_timestamp)
    for i in range(0, pos):
        instants.popleft()
    return len(instants)


def _response(status, body):
    body["status"] = "ok" if status == 200 else "fail"
    return web.Response(status=status, body=json.dumps(body))


class HealthcheckMonitored(ABC):
    @abstractmethod
    def healthcheck(self):
        pass


class Healthcheck(HttpHook):
    def __init__(
        self,
        barterdude: BarterDude,
        path: str = "/healthcheck",
        success_rate: float = 0.95,
        health_window: float = 60.0,  # seconds
        max_connection_fails: int = 3
    ):
        self.__barterdude = barterdude
        self.__success_rate = success_rate
        self.__health_window = health_window
        self.__success = deque()
        self.__fail = deque()
        self.__force_fail = False
        self.__connection_fails = 0
        self.__max_connection_fails = max_connection_fails
        super(Healthcheck, self).__init__(barterdude, path)

    def force_fail(self):
        self.__force_fail = True

    async def before_consume(self, message: RabbitMQMessage):
        pass

    async def on_success(self, message: RabbitMQMessage):
        self.__success.append(time())

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        self.__fail.append(time())

    async def on_connection_fail(self, error: Exception, retries: int):
        self.__connection_fails = retries

    async def __call__(self, req: web.Request):
        if self.__force_fail:
            return _response(500, {
                "message": "Healthcheck fail called manually"
            })

        if self.__connection_fails >= self.__max_connection_fails:
            return _response(500, {
                "message": (
                    "Reached max connection fails "
                    f"({self.__max_connection_fails})"
                )
            })

        response = {}
        status = 200

        all_monitored_modules_passed = True
        for module in self.__barterdude:
            if isinstance(self.__barterdude[module], HealthcheckMonitored):
                passed = self.__barterdude[module].healthcheck()
                response[module] = "ok" if passed else "fail"
                all_monitored_modules_passed &= passed

        if not all_monitored_modules_passed:
            status = 500

        old_timestamp = time() - self.__health_window
        success = _remove_old(self.__success, old_timestamp)
        fail = _remove_old(self.__fail, old_timestamp)

        if success == 0 and fail == 0:
            response["message"] = (
                f"No messages in last {self.__health_window}s"
            )
            return _response(status, response)

        rate = success / (success + fail)

        if rate < self.__success_rate:
            status = 500

        response.update({
            "message":
                f"Success rate: {rate} (expected: {self.__success_rate})",
            "fail": fail,
            "success": success
        })

        return _response(status, response)



import jsonschema
from functools import partial
from typing import Optional, Union
from asyncworker.rabbitmq.message import RabbitMQMessage
from asyncworker.easyqueue.exceptions import UndecodableMessageException


class ValidationException(jsonschema.ValidationError):
    pass


class Message:
    def __init__(self, message: RabbitMQMessage):
        self._message = message

    @property
    def body(self):
        return self._message.body

    @property
    def raw(self):
        return self._message.serialized_data

    @property
    def queue_name(self):
        return self._message._amqp_message.queue_name

    @property
    def properties(self):
        return self._message._amqp_message._properties

    @property
    def envelope(self):
        return self._message._amqp_message._envelope

    def accept(self):
        return self._message.accept()

    def reject(self, requeue=True):
        return self._message.reject(requeue)

    async def process_success(self):
        return await self._message.process_success()

    async def process_exception(self):
        return await self._message.process_exception()


class MessageValidation:
    def __init__(self, validation_schema: Optional[dict] = {}):
        self._validate = bool(validation_schema)
        resolver = jsonschema.RefResolver.from_schema(validation_schema)
        self._builder = partial(
            jsonschema.validate,
            schema=validation_schema,
            resolver=resolver
        )

    def validate(self, message: Union[RabbitMQMessage, Message]):
        try:
            body = message.body
            if self._validate:
                self._builder(body)
        except (jsonschema.ValidationError,
                UndecodableMessageException) as err:
            raise ValidationException(err)

    def __call__(self, message: RabbitMQMessage):
        self.validate(message)
        return Message(message)



import asyncio
from typing import Dict, Any, Iterable, Tuple
from collections.abc import MutableMapping
from aioamqp.properties import Properties


class ObjectWithCallsTracking:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method_calls = []

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)
        if not hasattr(attr, '__call__') or name == 'get_calls':
            return attr

        if asyncio.iscoroutine(attr) or asyncio.iscoroutinefunction(attr):
            async def async_track_calls(*args, **kwargs):
                self.method_calls.append((
                    name, args, kwargs
                ))
                return await attr(*args, **kwargs)
            return async_track_calls

        def track_calls(*args, **kwargs):
            self.method_calls.append((
                name, args, kwargs
            ))
            return attr(*args, **kwargs)
        return track_calls

    def get_calls(self):
        return [
            {'method': method, 'args': args, 'kwargs': kwargs}
            for method, args, kwargs in self.method_calls
        ]


class RabbitMQMessageMock(ObjectWithCallsTracking):
    def __init__(
        self,
        body: Dict = None,
        headers: Dict = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.body = body
        self.properties = Properties()
        self.properties.headers = headers

    def accept(self, *args, **kwargs):
        pass

    def reject(self, *args, **kwargs):
        pass


class BarterdudeMock(MutableMapping, ObjectWithCallsTracking):
    def __init__(
        self,
        mock_dependencies: Iterable[Tuple[Any, str]] = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__app = {}

        if mock_dependencies:
            for service, name in mock_dependencies:
                self[name] = service

    async def publish_amqp(self, *args, **kwargs):
        pass

    def __getitem__(self, key):
        return self.__app[key]

    def __setitem__(self, key, value):
        self.__app[key] = value

    def __delitem__(self, key):
        del self.__app[key]

    def __len__(self):
        return len(self.__app)

    def __iter__(self):
        return iter(self.__app)



import json
import traceback
from aiohttp import web
from asyncio import gather
from asyncworker import App, RouteTypes
from asyncworker.options import Options
from asyncworker.connections import AMQPConnection
from asyncworker.rabbitmq.message import RabbitMQMessage
from collections.abc import MutableMapping
from typing import Iterable, Optional, Callable, Any, Tuple
from barterdude.monitor import Monitor
from barterdude.message import MessageValidation, ValidationException
from barterdude.mocks import RabbitMQMessageMock, BarterdudeMock


class BarterDude(MutableMapping):
    def __init__(  # nosec
        self,
        hostname: str = "127.0.0.1",
        username: str = "guest",
        password: str = "guest",
        prefetch: int = 10,
        connection_name: str = "default",
    ):
        self.__connection = AMQPConnection(
            name=connection_name,
            hostname=hostname,
            username=username,
            password=password,
            prefetch=prefetch
        )
        self.__app = App(connections=[self.__connection])

    def add_endpoint(self, routes, methods, hook):
        self.__app.route(
            routes=routes,
            methods=methods,
            type=RouteTypes.HTTP
        )(hook)

    def add_callback_endpoint(
        self,
        routes: Iterable[str],
        hook: Callable,
        mock_dependencies: Iterable[Tuple[Any, str]] = None,
    ):
        async def hook_to_callback(req: web.Request):
            return await self._call_callback_endpoint(
                req,
                hook,
                mock_dependencies
            )

        self.add_endpoint(
            routes=routes,
            methods=['POST'],
            hook=hook_to_callback
        )

    async def _call_callback_endpoint(
        self,
        request: web.Request,
        hook: Callable,
        mock_dependencies: Iterable[Tuple[Any, str]],
    ):
        payload = await request.json()
        body = payload.get('body')
        headers = payload.get('headers')
        should_mock_barterdude = payload.get('should_mock_barterdude', True)

        if body is None:
            return web.Response(
                status=400,
                body=json.dumps({
                    'msg': 'Missing "body" attribute in payload.'
                })
            )

        rabbitmq_message_mock = RabbitMQMessageMock(body, headers)

        barterdude_mock = None
        if should_mock_barterdude:
            barterdude_mock = BarterdudeMock(mock_dependencies)

        response = {}

        try:
            await hook(rabbitmq_message_mock, barterdude=barterdude_mock)
        except Exception:
            response['exception'] = traceback.format_exc()

        response['message_calls'] = rabbitmq_message_mock.get_calls()

        if barterdude_mock is not None:
            response['barterdude_calls'] = barterdude_mock.get_calls()

        return web.Response(status=200, body=json.dumps(response))

    def consume_amqp(
        self,
        queues: Iterable[str],
        monitor: Monitor = Monitor(),
        coroutines: int = 10,
        bulk_flush_interval: float = 60.0,
        requeue_on_fail: bool = True,
        requeue_on_validation_fail: bool = False,
        validation_schema: Optional[dict] = {}
    ):
        msg_validation = MessageValidation(validation_schema)

        def decorator(f):
            async def process_message(message: RabbitMQMessage):
                await monitor.dispatch_before_consume(message)
                try:
                    await f(msg_validation(message))
                except ValidationException as error:
                    await monitor.dispatch_on_fail(message, error)
                    message.reject(requeue_on_validation_fail)
                except Exception as error:
                    await monitor.dispatch_on_fail(message, error)
                    message.reject(requeue_on_fail)
                else:
                    await monitor.dispatch_on_success(message)

            @self.__app.route(
                queues,
                type=RouteTypes.AMQP_RABBITMQ,
                options={
                    Options.BULK_SIZE: coroutines,
                    Options.BULK_FLUSH_INTERVAL: bulk_flush_interval,
                    Options.CONNECTION_FAIL_CALLBACK:
                        monitor.dispatch_on_connection_fail,
                }
            )
            async def wrapper(messages: RabbitMQMessage):
                await gather(*map(process_message, messages))

            return wrapper

        return decorator

    async def publish_amqp(
        self,
        exchange: str,
        data: dict,
        properties: dict = None,
        routing_key: str = "",
        **kwargs,
    ):
        await self.__connection.put(
            exchange=exchange,
            data=data,
            properties=properties,
            routing_key=routing_key,
            **kwargs
        )

    async def startup(self):
        await self.__app.startup()

    async def shutdown(self):
        await self.__app.shutdown()

    def run(self):
        self.__app.run()

    def __getitem__(self, key):
        return self.__app[key]

    def __setitem__(self, key, value):
        self.__app[key] = value

    def __delitem__(self, key):
        del self.__app[key]

    def __len__(self):
        return len(self.__app)

    def __iter__(self):
        return iter(self.__app)



from prometheus_client.metrics import MetricWrapperBase
from prometheus_client import (
    Counter,
    Gauge,
    Summary,
    Histogram,
    Info,
    Enum
)
from functools import partial


class Metrics(dict):

    def __init__(self, registry):
        self.__registry = registry

    @property
    def counter(self) -> Counter.__class__:
        return partial(Counter, registry=self.__registry)

    @property
    def gauge(self) -> Gauge.__class__:
        return partial(Gauge, registry=self.__registry)

    @property
    def summary(self) -> Summary.__class__:
        return partial(Summary, registry=self.__registry)

    @property
    def histogram(self) -> Histogram.__class__:
        return partial(Histogram, registry=self.__registry)

    @property
    def info(self) -> Info.__class__:
        return partial(Info, registry=self.__registry)

    @property
    def enum(self) -> Enum.__class__:
        return partial(Enum, registry=self.__registry)

    def __setitem__(self, name: str, metric: MetricWrapperBase):
        if name in self:
            value = self.__getitem__(name)
            raise ValueError(
                f"Key {name} already exists with value {value}"
            )
        super(Metrics, self).__setitem__(name, metric)



from setuptools import setup, find_packages

libs = []  # if need extra libs

extra = {
    "all": [],

}

with open("requirements/requirements_base.txt") as reqs:
    raw_requirements = reqs.read().split("\n")

requirements = []
for r in raw_requirements:
    if r.startswith("git+"):
        name = r.split(".git@")[0].split("/")[-1]
        requirements.append(f"{name} @ {r}")
    else:
        requirements.append(r)

for lib in libs:
    with open(f"requirements/requirements_{lib}.txt") as reqs:
        extra[lib] = reqs.read().split("\n")
        extra["all"] = extra["all"] + extra[lib]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="barterdude",
    description="Message exchange engine to build pipelines using brokers like RabbitMQ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Olx",
    license='Apache 2',
    include_package_data=True,
    url='https://github.com/olxbr/BarterDude/',
    download_url='https://github.com/olxbr/BarterDude/archive/master.zip',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=requirements,
    extras_require=extra,
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks"
    ],
    packages=find_packages()
)
